#! /usr/bin/perl
# access.pl
########################################################################

########################################################################
# ACCESS CONTROL HOOK FOR SUBVERSION
#
# Programmed by David Weintraub
# Date: 6-June-2005
# Revision: $Id: access.pl 256 2005-07-26 14:47:57Z weintraub $
# Purpose:
#    To control access to files and properties.
#
#    See comentary in control-file.template for more information.
#
# Note:
#    For debugging purposes, you can use the "-r" switch to run the
#    checks on an already existing revision. In this case, you could
#    call this program directory from the command line.
#
#    Also note that this program wont display any output -even from
#    the debug statements- unless pre-commit script returns a non-zero
#    exit status. In the pre-commit script, change the final statement
#    from "exit $?" to "exit 2". Remember to reset it or no one will be
#    able to commit anything.
#
########################################################################

########################################################################
# PERL MODULES
#
use Getopt::Long;
#
########################################################################

########################################################################
# CONSTANTS
#
*DEFAULT_CONTROL_FILE = \"./control-file";
   our $DEFAULT_CONTROL_FILE;
*DEFAULT_DEBUG_LEVEL = \0;
   our $DEFAULT_DEBUG_LEVEL;
*SVNLOOK_DEFAULT = \"/usr/bin/svnlook";
#*SVNLOOK_DEFAULT = \"/usr/local/bin/svnlook";
   our $SVNLOOK_DEFAULT;
#
########################################################################

########################################################################
# USAGE
#
our $USAGE = <<EOF;
usage:
    access.pl [-file <ctrlFile>] (-r<revision>|-t<transaction>) \\
        [-debug <debugLevel>] [-svnlook <svnlookCmd>] <repository>
    where:
	<ctrlFile>: Control File used for determining permissions
	    defaults to $DEFAULT_CONTROL_FILE.
	<revision>: Revision Number of archive (for testing)
	<transaction>: Transaction Number of Commit
	<debugLevel>: The debug level to implement>. Default
	    is $DEFAULT_DEBUG_LEVEL (zero means no debugging messages)
	<repository>: Full Path to Repository
	<svnlookCmd>: The svnlook command including the full path to
	              the command.
EOF
#
########################################################################

########################################################################
# PRAGMAS
#
use strict;
use warnings;
#
########################################################################

########################################################################
# OTHER VARIABLES
#
my %groupHash;	#Hash of Lists containing group definitions
my @propList;	#Array of Hash containing property definitions
my @revPropList; #Array of Hash containing rev property definitions
my @fileList;	#Array of Hash containing File Perm definitions
my @bannedFileList; #Array Containing Names of Banned Files
my $svnlookCmd;	#The full path and command of "svnlook"
my $lineNumber = 0;
my $author;	#Name of user who made the changes
my @errorList = ();	#List of parsing error from Control File
#
########################################################################

########################################################################
#  GET COMMAND LINE OPTIONS
#
my $transactionNum;
my $revisionNum;
my $controlFile;
my $option;
my $repository;
my $debugLevel;

GetOptions(
    "t=s" => \$transactionNum,
    "r=i" => \$revisionNum,
    "file=s" => \$controlFile,
    "debug=i" => \$debugLevel,
    "snvlook=s" => \$svnlookCmd
);

if ($revisionNum and $transactionNum) {
    die qq(You cannot specify both a revision and transaction number\n) .
	qq($USAGE\n);
}

unless ($controlFile) {
    $controlFile = $DEFAULT_CONTROL_FILE;
}

unless ($repository = $ARGV[0]) {
    die qq(You must specify a repository\n$USAGE\n);
}

unless ($svnlookCmd) {
    $svnlookCmd = $SVNLOOK_DEFAULT;
}

if ($revisionNum) {
    $option = "-r $revisionNum";
} elsif ($transactionNum) {
    $option = "-t $transactionNum";
} else {
    die qq(You must specify either the "-t" or "-r" option\n$USAGE\n);
}

unless ($debugLevel) {
    $debugLevel = $DEFAULT_DEBUG_LEVEL;
}

########################################################################
# SUBROUTINE DEBUG
#
# This routine is here for your own convinience. Sprinkle it liberally
# throughout this program if you need to debug something. The format
# is pretty simple:
#
# debug ("<debugStatement>", <level>");
#
#    Where:
#       <debugStatement> is a string to print out when statement executes
#       <level> is the debug level. The level is set when you execute the
#               hook. The statement won't print unless the -debug parameter
#               is equal to or higher than the <level>.
#
# My favorite trick is to simply take a statement and surround it with qq()
# quotes. This requires minimal retouching on my part. For example, I want
# to see what is happening with this statement:
#
#   if ($foo -eq $bar) {
#
# I will change that to this:
#
#   debug (qq(if ($foo -eq $bar) {/n), 2);
#
# Basically, I insert:
#
#   debug (qq(
#
# before the statement, and
#
#   \n), 2);
#
# After the statement. Sometimes, you want to disable a variable:
#
#    $result = $product / $divisor;
#
# I want to see what $product and $divisor are, but I really don't need
# to see "$result":
#
#    debug (qq(\$result = $product / $divisor;\n), 2);
#
# Note the backslash in front of $result.
#
sub debug {
    if (($debugLevel > 0) && ($_[1] <= $debugLevel)) {
	print "    " x ($_[1] - 1) if ($_[1] >= 1);
	print qq($_[0]\n);
    }
}
#
########################################################################

########################################################################
# SUBROUTINE ADD GROUP
#
# Purpose: To add a list of names to the groupHash (Which is a hash of lists)
#
sub addGroup {
    unless (/[^=]+\s*=\s*[^=]+/) {	#Check for Syntax Errors
	push (@errorList,
	    "Line $lineNumber: Error in parsing group:\n\t$_");
	return 1;
    }
    (my $group, my $members) = split (/\s*=\s*/);
    $group =~ s/\s//g;		#Remove Blanks
    $group =~ s/^@//;		#Remove "@" if on group
    $group = lc($group);	#Lowercase (Case is insignificant)
    foreach my $member (split(/\s*,\s*/, $members)) {
	$member =~ s/\s//g;
	$member = lc($member);	#Case is insignificant
	if($member =~ /^@/) {	#This is a group!
	    $member =~ s/^@//;	#Name of Group
	    if (defined($groupHash{$member})) {
		push(@{$groupHash{"$group"}}, @{$groupHash{"$member"}});
	    } else {
		push(@errorList,
		    qq(Line #$lineNumber: Group "$member" undefined));
	    }
	} else {
	    push(@{$groupHash{"$group"}}, "$member");
	}
    }
    return 0;
}
#
########################################################################

########################################################################
# SUBROUTINE ADD PROPERTY
#
sub addProp {
    unless (/(\s*[^:]+)\s*:\s*([^=]+)\s*=\s*(.*)/) {
	push(@errorList,
	    "Line $lineNumber: Error in parsing property:\n\t$_");
	return 1;
	}
	my $subscript = $#propList + 1;	#Next free entry in @propList
	my $filePattern = $1;
	my $property = $2;
	my $propValue = $3;
	$filePattern =~ s/\s*(\S*)\s*$/$1/;
	$property =~ s/\s//g;		#Remove Blanks

#
#   ####Check Property Type
#
	if ($propValue =~ /^\s*\//) {	#Regular Expression
	    $propList[$subscript]->{"type"} = "regex";
	    $propValue =~ s/^\s*\/(.*)\/\s*$/$1/;
	}
	elsif ($propValue =~ /^\s*"/) { #String
	    $propList[$subscript]->{"type"} = "string";
	    $propValue =~ s/^\s*"(.*)"\s*$/$1/;
	}
	elsif ($propValue =~ /^\d+$/) {
	    $propList[$subscript]->{"type"} = "number";
	    $propValue =~ s/^\s*(.*)\s*$/$1/;
	}
	else {
	    push(@errorList,
		qq(Line $lineNumber: Unknown prop type: "$propValue" ) .
		qq(Must be string (use quotes) regex (use "/") or ) .
		qq(number (digits only)));
	    return 1;
	}
	$propList[$subscript]->{"pattern"} = $filePattern;
	$propList[$subscript]->{"property"} = $property;
	$propList[$subscript]->{"value"} = $propValue;

    return 0;
}
#
########################################################################

########################################################################
# SUBROUTINE ADD REVISION PROPERTY
#
sub addRevProp {
    unless (/\s*([^=]+)\s*=\s*(.*)/) {
	push(@errorList,
	    "Line $lineNumber: Error in parsing revision property:\n\t$_");
	return 1;
	}
	my $subscript = $#revPropList + 1;	#Next free entry in @propList
	my $property = $1;
	my $propValue = $2;
	$property =~ s/\s//g;		#Remove Blanks

#
#   ####Check Property Type
#
	if ($propValue =~ /^\s*\//) {	#Regular Expression
	    $revPropList[$subscript]->{"type"} = "regex";
	    $propValue =~ s/^\s*\/(.*)\/\s*$/$1/;
	}
	elsif ($propValue =~ /^\s*"/) { #String
	    $revPropList[$subscript]->{"type"} = "string";
	    $propValue =~ s/^\s*"(.*)"\s*$/$1/;
	}
	elsif ($propValue =~ /^\d+$/) {
	    $revPropList[$subscript]->{"type"} = "number";
	    $propValue =~ s/^\s*(.*)\s*$/$1/;
	}
	else {
	    push(@errorList,
		qq(Line $lineNumber: Unknown prop type: "$propValue" ) .
		qq(Must be string (use quotes) regex (use "/") or number (digits only)));
	    return 1;
	}
	$revPropList[$subscript]->{"property"} = $property;
	$revPropList[$subscript]->{"value"} = $propValue;

    return 0;
}
#
########################################################################

########################################################################
# SUBROUTINE ADD FILE
#
# Purpose: To add a new file permission "record" to a List of Files
#    The record consists of a
#       * File Regex ("pattern")
#       * Permission ("permission") of either "a", "w", or "r"
#       * A hash of users with that permission ("member")
#
sub addFile {
    (my $filePattern, my $permission, my $members) = split(":");

    my $subscript = $#fileList + 1; 	#Subscript of fileList Array
    $filePattern =~ s/<USER>/$author/g;
    $fileList[$subscript]->{"pattern"} = $filePattern;
#
#   ####Check File Permissions
#

    if ($permission =~ /^(a|w|r)$/i) {
	$fileList[$subscript]->{"permission"} = lc($permission);
    } else {
	push(@errorList,
	    qq(Line #$lineNumber: Bad File Permission: "$permission"));
	return 1;
    }

#
#   ####Parse Group Permissions
#

    my @memberList = ();
    foreach my $member (split(/\s*,\s*/, $members)) {
	$member =~ s/\s//g;
	if ($member eq "\@ALL") { #Special Case: Line for All Users 
	    @memberList = ();
	    $memberList[0] = "\@ALL";
	    last;		#All are included. Nothing else to parse
	}
	$member = lc($member);	#Case is insignificant
	if($member =~ /^@/) {	#This is a group!
	    $member =~ s/^@//;	#Name of Group
	    if (defined($groupHash{$member})) {
		push(@memberList, @{$groupHash{"$member"}});
	    } else {
		push(@errorList,
		    qq(Line: $lineNumber Group "$member" underfined));
	    }
	} else {
	    push(@memberList, "$member");
	}
    }

#
#   ####Reformat member group into a hash

    foreach my $member (@memberList) {
	$fileList[$subscript]->{"member"}->{"$member"} = 1;
    }
    return 0;
}
#
########################################################################

########################################################################
# SUBROUTINE ADD BANNED FILE
#
# Purpose: To add a new file name "record" to a List of Files
#    that are not allowed to be committed.
#
#    The record consists of a
#       * File Regex ("pattern")
#
sub addBannedFile {
    (my $filePattern) = $_;

    my $subscript = $#bannedFileList + 1; 	#Subscript of fileList Array
    $bannedFileList[$subscript]->{"pattern"} = $filePattern;

    return 0;
}
#
########################################################################

########################################################################
# OPEN THE CONTROL FILE
#
open (CONTROL, "$controlFile") or
    die "Cannot open file \"$controlFile\" for reading\n";
select(STDERR);	#Only STDERR lines print out
#
########################################################################

########################################################################
# PARSE IN CONTROL FILE
#


#
#   ####Need Author because you need to replace <USER> in patterns
#

chomp ($author = qx($svnlookCmd author $option $repository));

while (<CONTROL>)
{
    $lineNumber++;
    chomp;
    next if (/^(#|--\s|\/\/)/);	#Ignore comments
    next if (/^\s*$/);		#Ignore Blank Lines
    if (s/group[^:]*://i){	# This is a Group Definition
	addGroup();
    }
    elsif(s/prop[^:]*://i) {	#This is a Property Definition
	addProp();
    }
    elsif(s/file[^:]*://i) {	#This is a File Definition
	addFile();
    }
    elsif(s/rev[^:]*://i) {	#This is a Revision Property Definition
	addRevProp();
    }
    elsif(s/ban[^:]*://i) {	#This is a Banned File
	addBannedFile();
    }
    else {			#This is an error!
	push (@errorList,
	   "Line $lineNumber: Bad syntax in $controlFile\n\t$_");
    }
}

#
#   ####If Errors in Control File, Stop Processing
#

if (@errorList) {
    print qq(Commit failed due to parsing errors in file "$controlFile"\n);
    for my $error (@errorList) {
	print "    $error\n";
    }
    print "Number of errors: " . scalar(@errorList) . "\n";
    exit 2;
}
#
########################################################################

########################################################################
# PRINT GROUPS
#
=pod
foreach my $groupName (sort keys (%groupHash)) {
    foreach my $member (@{$groupHash{"$groupName"}}) {
    }
}
#
########################################################################

########################################################################
# PRINT PROPERTIES
#
foreach my $property (@propList) {
	qq("$property->{value}"\n\n), 3);
}
#
########################################################################

########################################################################
# PRINT PROPERTIES
#
foreach my $property (@propList) {
	qq("$property->{value}"\n\n), 2);
}
#
########################################################################

########################################################################
# PRINT FILES
#
foreach my $file (@fileList) {
    foreach my $member (sort keys %{$file->{"member"}}) {
    }
}
=cut
#
########################################################################

########################################################################
# GET INFORMATION ABOUT THE FILES CHANGED
#

my $cmd = "$svnlookCmd changed $option $repository";
open (CHANGED, "$cmd|")
    or die qq(Cannot execute the command "$cmd"\n");

my @changeList = ();
my @checkPropList = ();
my @addList = ();

while (<CHANGED>)
{
    chomp;
    /([\S]+)\s+(.*)/;		#Can't use "split". Filename may have spaces
    my $status = $1;
    my $file = $2;

    if ($status =~ /^U/) {
	push (@changeList, $file);	#File Modified (Need "w" permission)
	push (@checkPropList, $file);	#Check Property
    }
    elsif ($status =~ /^A/) {
	push (@addList, "$file");		#File Added (Need "w" or "a" perm)
	push (@checkPropList, $file);	#Check Property
    }
    elsif ($status =~ /^D/) {
	push (@changeList, $file);	#File Modified (Need "w" permission)
    }
    elsif ($status =~ /^.U/)		#File Prop Modified!
    {
	push (@checkPropList, $file);
    }
    else
    {
	die qq(Can't interpret change: "$_"\n);
    }
}
#
########################################################################

########################################################################
# FOR EACH FILE, SEE IF YOU HAVE THE PERMISSION TO CHANGE THE FILE
#
# Flag Status:
#  -1: Not specified (Allowed)
#   1: Specified Yes (Allowed)
#   0: Specifed No (Not Allowed)
#
my @userRejectList = ();	#List of rejected files

#
#    ####Check for Modifications (User is either not specified or given "w")
#

foreach my $file (@changeList) {
my $userWriteFlag = -1;		#User Write Permission Not Specified
    foreach my $fileEntry (@fileList) {
	if ($file =~ /$fileEntry->{"pattern"}/) { #File matches pattern
	    if (($fileEntry->{"member"}->{"$author"}) or 
		($fileEntry->{"member"}->{"\@ALL"})) {
		if ($fileEntry->{"permission"} =~ /^w$/i) {
		    $userWriteFlag = 1;
		    last;		#No need to check farther
		}
		elsif ($userWriteFlag == -1) {	#If not already Set
		    $userWriteFlag = 0;	#User doesn't have permission (yet)
		}	#Checking whether user is "w", "a", or "r"
	    }	    #If Permission applies to user
	}	#If pattern matches
    }
    if ($userWriteFlag == 0) {
	push (@userRejectList, qq(No permission to change file "$file"));
    }
}

#
#   ####Check for Additions (User not specified or given "a" or "w")
#

foreach my $file (@addList) {

#
#   ####Checking for Add Permissions for Control-File "file:" Specs
#

my $userWriteFlag = -1;		#User Write Permission Not Specified
    foreach my $fileEntry (@fileList) {
	if ($file =~ /$fileEntry->{"pattern"}/) { #File matches pattern
	    if (($fileEntry->{"member"}->{"$author"}) or 
		($fileEntry->{"member"}->{"\@ALL"})) {
		if ($fileEntry->{"permission"} =~ /^[aw]$/i) {
		    $userWriteFlag = 1;
		    last;		#No need to check farther
		}
		elsif ($userWriteFlag == -1) {	#If not already Set
		    $userWriteFlag = 0;	#User doesn't have permission (yet)
		}	#Checking whether user is "w", "a", or "r"
	    }	    #If Permission applies to user
	}	#If pattern matches
    }	# For each entry in File Permission List

    if ($userWriteFlag == 0) {
	push (@userRejectList, qq(No permission to add file "$file"));
    }

#
#   ####Checking for Banned Files in Control-File "banned:" Specs
#

    $userWriteFlag = 1;		#Assume Permission is Granted
    foreach my $fileEntry (@bannedFileList) {
	if ($file =~ /$fileEntry->{"pattern"}/) { #File matches pattern
	    $userWriteFlag = 0;	#Banned File - Can't Commit!
	}
    }

    if ($userWriteFlag == 0) {
	push (@userRejectList, qq("$file" can't be added to respostory. ) .
	    qq(Banned!));
    }
}
#
########################################################################

########################################################################
# CHECK PROPERTIES
#
foreach my $file (@checkPropList) {
    foreach my $fileEntry(@propList) {
	if ($file =~ /$fileEntry->{"pattern"}/) { #Check This Property
	    my $property = $fileEntry->{"property"};
	    my $cmd = "$svnlookCmd propget $option " .
		"$repository $fileEntry->{property} $file 2> /dev/null";
	    unless (my $propValue = qx($cmd)) {
		push(@userRejectList,
		    qq(Property "$property" needed for file "$file".\n) .
		    qq(        Must match $fileEntry->{type} ) .
		    qq('$fileEntry->{value}'.));
	    } else {		#Property Found: Is it a good value?
		chomp ($propValue);
		if ($fileEntry->{"type"} eq "regex") {
		    if ($propValue !~ /$fileEntry->{"value"}/) {
			push (@userRejectList, 
			    qq(Property "$property" for file "$file" ) .
			    qq(is invalid.\n) .
			    qq(        Must match regex ) .
			    qq(/$fileEntry->{value}/.));
		    }
		}
		elsif ($fileEntry->{"type"} eq "string") {
		    if ($propValue ne $fileEntry->{"value"}) {
			push (@userRejectList, 
			    qq(Property "$property" for file "$file" ) .
			    qq( is invalid.\n) .
			    qq(        Must match string ) .
			    qq("$fileEntry->{value}".));
		    }
		}
		else {	#Assume it is a number
		    if ($propValue != $fileEntry->{"value"}) {
			push (@userRejectList, 
			    qq(Property "$property" for file "$file" ) .
			    qq(is invalid.\n) .
			    qq(        Must match number ) .
			    qq($fileEntry->{value}.));
		    }
		}
	    }
	}	#If @propList has this file for this property
    } #For each file entry found in @propList
}
#
########################################################################

########################################################################
# CHECK REVISION PROPERTIES
#
foreach my $propEntry (@revPropList) {
    my $cmd;
    my $property = $propEntry->{"property"};


#
#   ####Kludge: This is a kludge. The "-t" flag really doesn't work on
#       the "svn propget --revprop" command. However, it would be nice
#       to have a way to check the log message anyway, so I will simply
#       kludge the "svnlook propget --revprop svn:log" to "svnlook log".
#

    if (($property eq "svn:log") and ($option =~ /^-t/))
    {
	$cmd = "$svnlookCmd log $option $repository 2> /dev/null";
    }
    else
    {
	$cmd = "$svnlookCmd propget --revprop $option " . 
	    "$repository $property #2> /dev/null";
    }
    unless(my $propValue = qx($cmd)) {
	push (@userRejectList,
	qq(Revision Property "$property" must be set!\n) .
	    qq(        Must be a $propEntry->{type} set to ) .
	    qq("$propEntry->{value}".));
    } else {	#Revision Property does exist!
	chomp ($propValue);
	if ($propEntry->{"type"} eq "regex") {
	    if ($propValue !~ /$propEntry->{"value"}/) {
		push (@userRejectList,
		    qq(Revision Property "$property" is invalid.\n) .
		    qq(        Currently set to "$propValue".\n) .
		    qq(        Must match regex /$propEntry->{"value"}/.));
	    }
	}
	elsif ($propEntry->{"type"} eq "string") {
	    if ($propValue ne "$propEntry->{value}") {
		push (@userRejectList,
		    qq(Revision Property "$property" is invalid.\n) .
		    qq(        Currently set to "$propValue".\n) .
		    qq(        Must match string "$propEntry->{"value"}".));
	    }
	}
	else {	#Assume it's a number
	    if ($propValue != $propEntry->{"value"}) {
		push (@userRejectList,
		    qq(Revision Property "$property" is invalid.\n) .
		    qq(        Currently set to "$propValue".\n) .
		    qq(        Must match number "$propEntry->{"value"}".));
	    }
	}
    }
}
#
########################################################################   

########################################################################
# CHECK USER REJECT LIST
#
if(scalar @userRejectList) {
    print "ERROR: Commit failed for the following reasons:\n";
    foreach my $entry (@userRejectList) {
	print "    $entry\n";
    }
    exit 2;
} else {	#Commit is fine
    exit 0;
}
#
########################################################################
